//redKontroll2 version 1.0  fredrikolofsson.com
//GNU GPL v2 - See the OBDEV license for further details.  (http://www.obdev.at/avrusb/)

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "usbdrv.h"

static void hardwareInit(void) {
	
	//--set pins to output
	DDRB= 0b00110000;						//output pins for shiftin
	DDRD= 0b00000101;						//output pins for usb
	
	//--internal pull-ups for inputs
	PORTB= 0b00000111;						//pb3 is for reading shiftin (q8)
	PORTD= 0b11111010;
	
	//--setup adc
	//ADCSRA |= (1<<ADPS0);					//prescale 2
	//ADCSRA |= (1<<ADPS1);					//prescale 4
	//ADCSRA |= (1<<ADPS2);					//prescale 16
	ADCSRA |= ((1<<ADPS2)|(1<<ADPS0));		//prescale 32
	ADCSRA |= (1<<ADEN);					//enable adc
	
	//--reset usb
	uchar i= 20;
	while(i) {								//300ms
		_delay_ms(15);
		i--;
	}
	DDRD= 0;
}

static uchar reportBuffer[8];				/* buffer for HID reports */
static uchar idleRate;						/* in 4 ms units */
uchar changed= 255;							//flag any data changed
uchar shiftIn(void);						//read data from 4021

PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = { //USB report descriptor
	0x05, 0x01,        // USAGE_PAGE (Generic Desktop)
	0x09, 0x04,        // USAGE (Joystick)
	0xa1, 0x01,        // COLLECTION (Application)
	
	0x09, 0x30,        // USAGE (adc0)
	0x15, 0x00,        // LOGICAL_MINIMUM (0)
	0x26, 0xff, 0x00,  // LOGICAL_MAXIMUM (255)
	0x75, 0x08,        // REPORT_SIZE (8)
	0x95, 0x01,        // REPORT_COUNT (1)
	0x81, 0x02,        // INPUT (Data,Var,Abs)
	
	0x09, 0x31,        // USAGE (adc1)
	0x15, 0x00,        // LOGICAL_MINIMUM (0)
	0x26, 0xff, 0x00,  // LOGICAL_MAXIMUM (255)
	0x75, 0x08,        // REPORT_SIZE (8)
	0x95, 0x01,        // REPORT_COUNT (1)
	0x81, 0x02,        // INPUT (Data,Var,Abs)
	
	0x09, 0x32,        // USAGE (adc2)
	0x15, 0x00,        // LOGICAL_MINIMUM (0)
	0x26, 0xff, 0x00,  // LOGICAL_MAXIMUM (255)
	0x75, 0x08,        // REPORT_SIZE (8)
	0x95, 0x01,        // REPORT_COUNT (1)
	0x81, 0x02,        // INPUT (Data,Var,Abs)
	
	0x09, 0x33,        // USAGE (adc3)
	0x15, 0x00,        // LOGICAL_MINIMUM (0)
	0x26, 0xff, 0x00,  // LOGICAL_MAXIMUM (255)
	0x75, 0x08,        // REPORT_SIZE (8)
	0x95, 0x01,        // REPORT_COUNT (1)
	0x81, 0x02,        // INPUT (Data,Var,Abs)
	
	0x09, 0x34,        // USAGE (adc4)
	0x15, 0x00,        // LOGICAL_MINIMUM (0)
	0x26, 0xff, 0x00,  // LOGICAL_MAXIMUM (255)
	0x75, 0x08,        // REPORT_SIZE (8)
	0x95, 0x01,        // REPORT_COUNT (1)
	0x81, 0x02,        // INPUT (Data,Var,Abs)
	
	0x09, 0x35,        // USAGE (adc5)				//7bits adc only. msb is used for digital input
	0x15, 0x00,        // LOGICAL_MINIMUM (0)
	0x26, 0xff, 0x00,  // LOGICAL_MAXIMUM (255)
	0x75, 0x08,        // REPORT_SIZE (8)
	0x95, 0x01,        // REPORT_COUNT (1)
	0x81, 0x02,        // INPUT (Data,Var,Abs)
	
	0x09, 0x36,        // USAGE (dig0)
	0x15, 0x00,        // LOGICAL_MINIMUM (0)
	0x26, 0xff, 0x00,  // LOGICAL_MAXIMUM (255)
	0x75, 0x08,        // REPORT_SIZE (8)
	0x95, 0x01,        // REPORT_COUNT (1)
	0x81, 0x02,        // INPUT (Data,Var,Abs)
	
	0x09, 0x37,        // USAGE (dig1)
	0x15, 0x00,        // LOGICAL_MINIMUM (0)
	0x26, 0xff, 0x00,  // LOGICAL_MAXIMUM (255)
	0x75, 0x08,        // REPORT_SIZE (8)
	0x95, 0x01,        // REPORT_COUNT (1)
	0x81, 0x02,        // INPUT (Data,Var,Abs)
	
	0xc0               // END_COLLECTION
};

static void buildReport(uchar adc0, uchar adc1, uchar adc2, uchar adc3, uchar adc4, uchar adc5, uchar dig0, uchar dig1) {
	reportBuffer[0]= adc0;
	reportBuffer[1]= adc1;
	reportBuffer[2]= adc2;
	reportBuffer[3]= adc3;
	reportBuffer[4]= adc4;
	reportBuffer[5]= adc5;
	reportBuffer[6]= dig0;
	reportBuffer[7]= dig1;
}

uchar usbFunctionSetup(uchar data[8]) {
	usbRequest_t *rq = (void *)data;

    usbMsgPtr = reportBuffer;
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
            /* we only have one report type, so don't look at wValue */
            buildReport(0, 0, 0, 0, 0, 0, 0, 0);
            return sizeof(reportBuffer);
        }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
            usbMsgPtr = &idleRate;
            return 1;
        }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
            idleRate = rq->wValue.bytes[1];
        }
    }else{
        /* no vendor specific requests implemented */
    }
	return 0;
}

int main(void) {
	
	uchar adc0= 0, adc1= 0, adc2= 0, adc3= 0, adc4= 0, adc5= 0, dig0= 0, dig1= 0;
	
	wdt_enable(WDTO_2S);
	hardwareInit();
	usbInit();
	sei();
	
	//--main loop
	while(1) {
		
		wdt_reset();
		usbPoll();
		
		//--read analog inputs
		ADMUX= 0b11000000;			//select adc channel 0
		ADCSRA |= (1<<ADSC);		//start conversion
		while((ADCSRA&64)==64);		//wait for ADSC bit to go low
		adc0= ADCL>>2;				//skip 2 lowest bits
		adc0 |= (ADCH<<6);
		if(reportBuffer[0]!=adc0) {
			changed= 0;
		}
		ADMUX= 0b11000001;			//select adc channel 1
		ADCSRA |= (1<<ADSC);		//start conversion
		while((ADCSRA&64)==64);		//wait for ADSC bit to go low
		adc1= ADCL>>2;				//skip 2 lowest bits
		adc1 |= (ADCH<<6);
		if(reportBuffer[1]!=adc1) {
			changed= 1;
		}
		ADMUX= 0b11000010;			//select adc channel 2
		ADCSRA |= (1<<ADSC);		//start conversion
		while((ADCSRA&64)==64);		//wait for ADSC bit to go low
		adc2= ADCL>>2;				//skip 2 lowest bits
		adc2 |= (ADCH<<6);
		if(reportBuffer[2]!=adc2) {
			changed= 2;
		}
		ADMUX= 0b11000011;			//select adc channel 3
		ADCSRA |= (1<<ADSC);		//start conversion
		while((ADCSRA&64)==64);		//wait for ADSC bit to go low
		adc3= ADCL>>2;				//skip 2 lowest bits
		adc3 |= (ADCH<<6);
		if(reportBuffer[3]!=adc3) {
			changed= 3;
		}
		ADMUX= 0b11000100;			//select adc channel 4
		ADCSRA |= (1<<ADSC);		//start conversion
		while((ADCSRA&64)==64);		//wait for ADSC bit to go low
		adc4= ADCL>>2;				//skip 2 lowest bits
		adc4 |= (ADCH<<6);
		if(reportBuffer[4]!=adc4) {
			changed= 4;
		}
		ADMUX= 0b11000101;			//select adc channel 5 (7bits adc only)
		ADCSRA |= (1<<ADSC);		//start conversion
		while((ADCSRA&64)==64);		//wait for ADSC bit to go low
		adc5= ADCL>>3;				//skip 3 lowest bits
		adc5 |= (ADCH<<5);
		if((PIND&(1<<1)) == (1<<1)) {
			adc5 |= (1<<7);
		}
		if(reportBuffer[5]!=adc5) {
			changed= 5;
		}
		
		//--read digital inputs from shiftIn
		dig0= shiftIn();
		if(reportBuffer[6]!=dig0) {
			changed= 6;
		}
		
		//--read digital inputs directly from pins
		if((PIND&(1<<3)) == (1<<3)) {
			dig1 |= (1<<0);
		} else {
			dig1 &= ~(1<<0);
		}
		if((PIND&(1<<4)) == (1<<4)) {
			dig1 |= (1<<1);
		} else {
			dig1 &= ~(1<<1);
		}
		if((PIND&(1<<5)) == (1<<5)) {
			dig1 |= (1<<2);
		} else {
			dig1 &= ~(1<<2);
		}
		if((PIND&(1<<6)) == (1<<6)) {
			dig1 |= (1<<3);
		} else {
			dig1 &= ~(1<<3);
		}
		if((PIND&(1<<7)) == (1<<7)) {
			dig1 |= (1<<4);
		} else {
			dig1 &= ~(1<<4);
		}
		if((PINB&(1<<0)) == (1<<0)) {
			dig1 |= (1<<5);
		} else {
			dig1 &= ~(1<<5);
		}
		if((PINB&(1<<1)) == (1<<1)) {
			dig1 |= (1<<6);
		} else {
			dig1 &= ~(1<<6);
		}
		if((PINB&(1<<2)) == (1<<2)) {
			dig1 |= (1<<7);
		} else {
			dig1 &= ~(1<<7);
		}
		if(reportBuffer[7]!=dig1) {
			changed= 7;
		}
		
		//--send report
		if(usbInterruptIsReady()&&(changed!=255)) {
			buildReport(adc0, adc1, adc2, adc3, adc4, adc5, dig0, dig1);
			usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
			changed= 255;
		}
	}
	return 0;
}

uchar shiftIn(void) {
	char i;
	uchar data= 0;
	PORTB |= (1<<5);					//set latch pin to collect parallel data
	_delay_us(20);
	PORTB &= ~(1<<5);					//clear latch pin to transmit
	for(i= 7; i>=0; i--) {
		PORTB &= ~(1<<4);				//clear clock pin
		if((PINB&(1<<3)) == (1<<3)) {	//read from pin
			data |= (1<<i);
		}
		PORTB |= (1<<4);				//set clock pin
	}
	return data;
}
